/* This file is part of toroidtoe, a web server implementing a game of
 * five-in-a-row on a grid warped onto a surface of a torus.  This
 * program is intended as a basis for student exercises on courses at
 * Helsinki University of Technology.  This source code is distributed
 * under the GPL.
 *
 * Author: Kenneth Oksanen <cessu@iki.fi>
 */

/* This file implements the artificial intelligence (AI) part of the
   problem: choosing the next move and learning from player games.

   Algorithms implemented in this file are rather complex, and we
   refer to the AI and computer games litterature if you wish to
   understand them - that won't be necessary for passing the course.
   In addition to standard text books and some web pages
	http://www.cs.biu.ac.il/~davoudo/intro.html
   that cover, for example alpha-beta-search, we refer to Aske Plaat's
   web page (and his articles) on MTD(f) search:
	http://home.tiscali.nl/askeplaat/mtdf.html
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <inttypes.h>

#include "ai.h"
#include "game.h"


#define INFINITY  6001

/* There are two kinds of formats to store a game position: the
   `position_t' is an NX*NY array of content_t's and intended for
   convenient and efficient access, while `compressed_position_t' is a
   much more concise but slower format for storing as many game
   positions as possible in the database and hash tables.  The
   compressed position uses cumbersome bit manipulations to store the
   position in typically one 16'th of the uncompressed `position_t'. */

typedef struct {
#define N_WORDS  (((NX) * (NY) + 15) / 16)
  uint32_t content[N_WORDS];
  value_t lower_bound, upper_bound;
} compressed_position_t;


/* Store the given content into the given place in the given
   compressed_position_t. */

static void compressed_position_set(compressed_position_t *cp,
				    int x, int y, content_t content)
{
  int xy;
  uint32_t c;

  assert(x >= 0);
  assert(x < NX);
  assert(y >= 0);
  assert(y < NY);

  xy = x * NY + y;
  c = cp->content[xy / 16];
  c &= ~(0x3 << ((xy % 16) * 2));
  c |= ((uint32_t) content) << ((xy % 16) * 2);
  cp->content[xy / 16] = c;
}


/* Returns non-zero if the two given compressed positions are equal.
   Other fields (lower_bound, upper_bound) are ignored. */

static int compressed_positions_are_identical(const compressed_position_t *cp1,
					      const compressed_position_t *cp2)
{
  return memcmp((cp1), (cp2), sizeof(uint32_t) * N_WORDS) == 0;
}


#ifdef DEBUG

/* These routines have been useful when debugging.
 */

static void compressed_position_print(const compressed_position_t *cp)
{
  int x, y, xy;

  for (y = NY - 1; y >= 0; y--) {
    for (x = 0; x < NX; x++) {
      xy = x * NY + y;
      printf(" %c", " OX?"[(cp->content[xy / 16] >> ((xy % 16)) * 2) & 0x3]);
    }
    printf("\n");
  }
}


static void position_print(const position_t p)
{
  int x, y;

  for (y = NY - 1; y >= 0; y--) {
    for (x = 0; x < NX; x++)
      printf(" %c", " OX?"[p[x][y]]);
    printf("\n");
  }
}

#endif


/* Positional evaluation at the leaves of the search tree, and also
   applied to improve the move orderig heuristic.  This is done with
   the help of a precomputed table represented with plain C in file
   `patterns.def' and included as a constant data below.  The table is
   indexed with a value obtained by encoding lines of eleven pieces in
   all four directions which as not reverses of eachother. */

const struct {
  signed int value : 14;
  unsigned int forced : 2;
} encoded_patterns[] = {
#include "patterns.h"
};


/* This function returns an unsigned integer which encodes the 11-place
   line in the given direction around the given place and position. */

static uint32_t encode_line(position_t p, int x, int y, int dx, int dy)
{
  uint32_t encoding = 0;
  int i;

  x -= 6 * dx;
  y -= 6 * dy;
  for (i = 0; i < 11; i++) {
    x += dx;
    y += dy;
    /* Take into account warping around edges of the array as if the
       array formed a torus. */
    if (x < 0)
      x += NX;
    else if (x >= NX)
      x -= NX;
    if (y < 0)
      y += NY;
    else if (y >= NY)
      y -= NY;
    encoding = N_CONTENTS * encoding + p[x][y];
  }
  return encoding;
}


/* The estimated value of the position is the sum of all 11-piece
   lines around non-empty places in the playing area.  This could be
   computed more efficiently incrementally during search, but we don't
   do it for sake of simplicity. */

static value_t evaluate_position(position_t p)
{
  int x, y;
  value_t value = 0;

  for (x = 0; x < NX; x++)
    for (y = 0; y < NY; y++)
      if (p[x][y] != EMPTY)
	value += encoded_patterns[encode_line(p, x, y, 0, 1)].value
	  + encoded_patterns[encode_line(p, x, y, 1, 1)].value
	  + encoded_patterns[encode_line(p, x, y, 1, 0)].value
	  + encoded_patterns[encode_line(p, x, y, 1, -1)].value;
  return value;
}


/* A routine which determines whether either side has won, or if the
   game is drawn.  Returns non-zero if so, and stores the outcome
   (from cross' point of view) into *value_p. */

int game_is_over(position_t p, value_t *value_p)
{
  int x, y, found_empty_places = 0;
  value_t v;

  for (x = 0; x < NX; x++)
    for (y = 0; y < NY; y++)
      if (p[x][y] != EMPTY) {
	v = encoded_patterns[encode_line(p, x, y, 0, 1)].value;
	if (v == WIN || v == -WIN) {
	  *value_p = v;
	  return 1;
	}
	v = encoded_patterns[encode_line(p, x, y, 1, 1)].value;
	if (v == WIN || v == -WIN) {
	  *value_p = v;
	  return 1;
	}
	v = encoded_patterns[encode_line(p, x, y, 1, 0)].value;
	if (v == WIN || v == -WIN) {
	  *value_p = v;
	  return 1;
	}
	v = encoded_patterns[encode_line(p, x, y, 1, -1)].value;
	if (v == WIN || v == -WIN) {
	  *value_p = v;
	  return 1;
	}
      } else
	found_empty_places = 1;
  if (!found_empty_places) {
    *value_p = DRAW;
    return 1;
  }
  return 0;
}


/* Hash tables.  There are two kinds of these: the first, called
   merely `hash_table' is used during . */

#define HASH_TABLE_SIZE   (1024*1024*4)

static compressed_position_t hash_table[HASH_TABLE_SIZE];
static compressed_position_t rote_learning_hash_table[HASH_TABLE_SIZE];

static uint32_t compressed_position_hash(const compressed_position_t *cp)
{
  uint32_t w, h = 0xDEADBEEF;
  int i;

  for (i = 0; i < N_WORDS; i++) {
    h -= w = cp->content[i];
    h += h << 10;
    h ^= h >> 28;
    h += w >> 11;
    h += h << 13;
    h ^= h >> 20;
    h += w >> 22;
    h += h << 13;
    h ^= h >> 3;
  }
  return h;
}


/* Take the verbose representation `p' of the position and encode it
   into a compressed position, into `*cp'.  Perform various
   canonifications in the process: move the origo of the coordinates,
   flip the direction of X- and Y-axes, and transpose, should the
   resulting form be deemed more "canonic" according to some
   heuristics. */

static void position_compress(position_t p, compressed_position_t *cp)
{
  int x, y, new_x_origo, new_y_origo, dx, dy, x2, y2;
  int hash_x[NX], hash_y[NY];
  int hash_value[/* content_t */] = { 0, 1, NX + NY, }, pv;

  memset(cp->content, 0, sizeof(uint32_t) * N_WORDS);
  memset(hash_x, 0, sizeof(hash_x));
  memset(hash_y, 0, sizeof(hash_y));

  /* Determine a "hash"-value for each column and row. */
  for (x = 0; x < NX; x++)
    for (y = 0; y < NY; y++) {
      pv = hash_value[p[x][y]];
      hash_x[x] += pv;
      hash_y[y] += pv;
    }

  /* Determine new origo of the X-coordinate so that the zero-column's
     hash value is maximal and direction of the X axis with a similar
     hash-value based heuristic. */
  for (new_x_origo = 0, x = 1; x < NX; x++)
    if (hash_x[x] > hash_x[new_x_origo])
      new_x_origo = x;
  if (hash_x[(new_x_origo + 1) % NX] > hash_x[(new_x_origo + NX - 1) % NX])
    dx = 1;
  else
    dx = -1;
  /* Same for the Y axis. */
  for (new_y_origo = 0, y = 1; y < NY; y++)
    if (hash_y[y] > hash_y[new_y_origo])
      new_y_origo = y;
  if (hash_y[(new_y_origo + 1) % NY] > hash_y[(new_y_origo + NY - 1) % NY])
    dy = 1;
  else
    dy = -1;

  /* Finally, move the board. */
#if NX == NY 
  /* Possibly transpose the board as well, but only if the board is as
     wide as it is high. */
  if (hash_x[new_x_origo] > hash_y[new_y_origo])
    for (x = 0, y2 = new_y_origo; x < NX; x++) {
      for (y = 0, x2 = new_x_origo; y < NY; y++) {
	compressed_position_set(cp, x, y, p[x2][y2]);
	x2 += dx;
	if (x2 < 0)
	  x2 = NX - 1;
	else if (x2 >= NX)
	  x2 = 0;
      }
      y2 += dy;
      if (y2 < 0)
	y2 = NY - 1;
      else if (y2 >= NY)
	y2 = 0;
    }
  else
#endif
    for (x = 0, x2 = new_x_origo; x < NX; x++) {
      for (y = 0, y2 = new_y_origo; y < NY; y++) {
	compressed_position_set(cp, x, y, p[x2][y2]);
	y2 += dy;
	if (y2 < 0)
	  y2 = NY - 1;
	else if (y2 >= NY)
	  y2 = 0;
      }
      x2 += dx;
      if (x2 < 0)
	x2 = NX - 1;
      else if (x2 >= NX)
	x2 = 0;
    }
}


/* Some utilities for move ordering in the `alpha_beta_with_memory'. */

typedef struct {
  int x, y, priority, forced, wins;
} move_t;


static int cmp_move_priority(const void *a, const void *b)
{
  const move_t *ma = (const move_t *) a;
  const move_t *mb = (const move_t *) b;

  if (ma->priority > mb->priority)
    return -1;
  else if (ma->priority < mb->priority)
    return 1;
  else
    return 0;
}


static uint32_t priorities[NX][NY];

#define MARK_KILLER(x, y)			\
  do {						\
    if (priorities[x][y] < 0xFFFF)		\
      priorities[x][y] += 10;			\
  } while (0)
#define UNMARK_KILLER(x, y)			\
  do {						\
    if (priorities[x][y] > 0)			\
      priorities[x][y] -= 10;			\
  } while (0)


/* `alpha_beta_with_memory' stores the best move at root to these
   variables. */

static int best_x, best_y;


/* The main tree-searching function. */

static value_t alpha_beta_with_memory(position_t p, content_t turn,
				      value_t alpha, value_t beta,
				      int effort, int depth)
{
  int x, y, i, ix, n_moves, forced, next_effort, have_empty_places;
  int have_forced_defences;
  value_t v, value, a, b, upper_bound, lower_bound;
  position_t new_p;
  compressed_position_t cp, *q;
  move_t moves[NX * NY + 1];
  uint32_t encoding_1, encoding_2, encoding_3, encoding_4;

  position_compress(p, &cp);
  ix = compressed_position_hash(&cp) % HASH_TABLE_SIZE;

  /* First look into the rote learning hash table, or "opening book",
     evaluated as root in previous games. */
  q = &rote_learning_hash_table[ix];
  if (depth > 0 && compressed_positions_are_identical(&cp, q)) {
    value = q->lower_bound;
    if (value >= WIN - NX * NY)
      value -= depth;
    else if (value <= - WIN + NX * NY)
      value += depth;
    return value;
  }

  /* Look into the hash table generated durig this search. */
  q = &hash_table[ix];
  if (compressed_positions_are_identical(&cp, q)) {
    lower_bound = q->lower_bound;
    upper_bound = q->upper_bound;
    if (lower_bound >= beta)
      return lower_bound;
    if (upper_bound <= alpha)
      return upper_bound;
    if (lower_bound > alpha)
      alpha = lower_bound;
    if (upper_bound < beta)
      beta = upper_bound;
  } else {
    lower_bound = -INFINITY;
    upper_bound = INFINITY;
  }

  /* A simple move ordering heuristic.  Alpha-beta search is
     significantly faster only if at each node of the tree the best
     move is searched as early as possible.  To accomplish this we
     implement three heuristics according to which we sort the moves:
       - (UN)MARK_KILLER and the priorities table implements something
         similar to killer and history heuristics.
       - The contribution of the move to the positional evaluation
       - Moves the patterns define as "forced", such as playing the
         fifth into a row. */
  have_forced_defences = have_empty_places = n_moves = 0;
  for (x = 0; x < NX; x++)
    for (y = 0; y < NY; y++)
      if (p[x][y] == EMPTY) {
	have_empty_places = 1;
	p[x][y] = turn;
	encoding_1 = encode_line(p, x, y, 0, 1);
	encoding_2 = encode_line(p, x, y, 1, 1);
	encoding_3 = encode_line(p, x, y, 1, 0);
	encoding_4 = encode_line(p, x, y, 1, -1);
	p[x][y] = EMPTY;
	forced = encoded_patterns[encoding_1].forced
	  || encoded_patterns[encoding_2].forced
	  || encoded_patterns[encoding_3].forced
	  || encoded_patterns[encoding_4].forced;
	if (encoded_patterns[encoding_1].forced == 2
	    || encoded_patterns[encoding_2].forced == 2
	    || encoded_patterns[encoding_3].forced == 2
	    || encoded_patterns[encoding_4].forced == 2) {
	  have_forced_defences = 1;
	  forced = 2;
	}
	if (effort > 0 || forced) {
	  moves[n_moves].x = x;
	  moves[n_moves].y = y;
	  moves[n_moves].forced = forced;
	  value = abs(encoded_patterns[encoding_1].value)
	    + abs(encoded_patterns[encoding_2].value)
	    + abs(encoded_patterns[encoding_3].value)
	    + abs(encoded_patterns[encoding_4].value);
	  moves[n_moves].priority =
	    priorities[x][y] + value + (forced ? 1000 : 0);
	  moves[n_moves].wins = (value >= WIN);
	  n_moves++;
	}
      }
  if (depth != 0 && (n_moves == 0 || !have_forced_defences)) {
    /* If deeper searching checks only threatening and forced replies,
       then add a "move" to place -1,-1 to indicate that a static
       evaluation of the position should also be regarded in the
       minimaxing. */
    moves[n_moves].x = moves[n_moves].y = -1;
    moves[n_moves].priority = moves[n_moves].wins = 0;
    n_moves++;
  }
  qsort(moves, n_moves, sizeof(move_t), cmp_move_priority);

  /* Below the core alpha-beta-logic as described in text books. */
  if (!have_empty_places)
    /* Nowhere to play, game ended in a draw. */
    value = 0;
  else if (turn == CROSS) {
    value = -INFINITY;
    a = alpha;
    for (i = 0; i < n_moves; i++) {
      x = moves[i].x;
      y = moves[i].y;
      if (x == -1)
	v = evaluate_position(p);
      else {
	memcpy(new_p, p, sizeof(position_t));
	new_p[x][y] = CROSS;
	if (moves[i].wins)
	  /* The depth term below tries to force winning play ASAP. */
	  v = WIN - depth;
	else {
	  next_effort = effort - 2 + moves[i].forced;
	  if (next_effort < 0)
	    next_effort = 0;
	  v = alpha_beta_with_memory(new_p, NOUGHT, a, beta,
				     next_effort, depth + 1);
	}
      }
      if (v > value) {
	value = v;
	if (depth == 0) {
	  best_x = x;
	  best_y = y;
	}
	if (value >= beta) {
	  MARK_KILLER(x, y);
	  goto out;
	}
	if (value > a)
	  a = value;
	else
	  UNMARK_KILLER(x, y);
      } else
	UNMARK_KILLER(x, y);
    }
  } else {
    /* This is very similar to the above variation, except the
       "NOUGHT" player tries to minimize the value and evaluates the
       position as -WIN if it wins. */
    assert(turn == NOUGHT);
    value = INFINITY;
    b = beta;
    for (i = 0; i < n_moves; i++) {
      x = moves[i].x;
      y = moves[i].y;
      if (x == -1)
	v = evaluate_position(p);
      else {
	memcpy(new_p, p, sizeof(position_t));
	new_p[x][y] = NOUGHT;
	if (moves[i].wins)
	  v = -WIN + depth;
	else {
	  next_effort = effort - 2 + moves[i].forced;
	  if (next_effort < 0)
	    next_effort = 0;
	  v = alpha_beta_with_memory(new_p, CROSS, alpha, b,
				     next_effort, depth + 1);
	}
      }
      if (v < value) {
	value = v;
	if (depth == 0) {
	  best_x = x;
	  best_y = y;
	}
	if (value <= alpha) {
	  MARK_KILLER(x, y);
	  goto out;
	}
	if (value < b)
	  b = value;
	else
	  UNMARK_KILLER(x, y);
      } else
	UNMARK_KILLER(x, y);
    }
  }

 out:
  /* Finally, store the position into the transposition hash table and
     update the known upper/lower bound of its value accordingly. */
  if (value <= alpha)
    upper_bound = value;
  else if (value > alpha && value < beta)
    upper_bound = lower_bound = value;
  else if (value >= beta)
    lower_bound = value;
  *q = cp;
  q->lower_bound = lower_bound;
  q->upper_bound = upper_bound;
  
  return value;
}


/* This function implements the MTD(f) driver and performs also the
   rote learning memoization once the search has completed.  This code
   does not use iterative deepening. */

value_t ai_move(position_t p, content_t turn,
		int *px, int *py, int level)
{
  value_t guess = DRAW, lower_bound = -INFINITY, upper_bound = INFINITY, beta;
  int x, y, n_moves, effort, i;

  /* Some preparations before the MTD(f) loop. */

  memset(&priorities, 0, sizeof(priorities));
  memset(hash_table, 0xFF, sizeof(hash_table));

  /* Count how many moves there are left. */
  n_moves = 0;
  for (x = 0; x < NX; x++)
    for (y = 0; y < NY; y++)
      if (p[x][y] == EMPTY)
	n_moves++;
  /* Compute the approximate search depth.  The formula below allows
     roughly 2^(level+8) nodes in the search tree.  This is only a
     very approximative formula, based on experiments and not on any
     mathematical property of the search space. */
  for (i = n_moves / 4, effort = 1;
       i < (1 << (level + 8)) && effort < n_moves;
       i *= (n_moves - effort++ + 3) / 4)
    ;

  /* Actual MTD(f) logic.
   */
  do {
    if (guess == lower_bound)
      beta = guess + 1;
    else
      beta = guess;
    best_x = best_y = -1;
    guess = alpha_beta_with_memory(p, turn, beta - 1, beta, effort, 0);
    if (guess < beta) {
      upper_bound = guess;
      if (turn == NOUGHT) {
	*px = best_x;
	*py = best_y;
      }
    } else {
      lower_bound = guess;
      if (turn == CROSS) {
	*px = best_x;
	*py = best_y;
      }
    }
  } while (lower_bound < upper_bound);
  
  if (n_moves > 6) {
    /* Store position into the rote learning hash table.  Positions
       very late in the game are not stored because they are rare and
       can typically searched exhaustively. */
    compressed_position_t cp;
    int i;
    
    position_compress(p, &cp);
    cp.lower_bound = cp.upper_bound = guess;
    i = compressed_position_hash(&cp) % HASH_TABLE_SIZE;
    rote_learning_hash_table[i] = cp;
  }

  return guess;
}


void ai_init(void)
{
  memset(rote_learning_hash_table, 0xFF, sizeof(rote_learning_hash_table));
}
