/* This file is part of toroidtoe, a web server implementing a game of
 * five-in-a-row on a grid warped onto a surface of a torus.  This
 * program is intended as a basis for student exercises on courses at
 * Helsinki University of Technology.  This source code is distributed
 * under the GPL.
 *
 * Author: Kenneth Oksanen <cessu@iki.fi>
 */


/* Networking code, a very conventional UNIX listening socket creation
   and a select-loop. */

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include "net.h"


/* Maximum size of messages that can be treated with one callback.
 */
#define BUF_SIZE  2048


void net_event_loop(callback_t cb)
{
  struct sockaddr_in s_in;
  int one = 1, sock, fd, max_fd, n_bytes;
  char buf[BUF_SIZE + 1];
  fd_set fds, read_fds;

  /* Create a socket, set suitable port and options for it, and listen
     to incoming connections. */

  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    perror("net_event_loop/socket");
    return;
  }
  if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
		 (char *) &one, sizeof(one)) < 0) {
    perror("net_event_loop/setsockopt");
    return;
  }

  s_in.sin_port = htons((u_short) 6666);
  s_in.sin_family = AF_INET;
  s_in.sin_addr.s_addr = INADDR_ANY;
  if (bind(sock, (struct sockaddr *) &s_in, sizeof(s_in)) < 0) {
    perror("net_event_loop/bind");
    close(sock);
    return;
  }
  if (listen(sock, 5)) {
    perror("net_event_loop/listen");
    close(sock);
    return;
  }

  /* Initializations before the main event loop. */

  FD_ZERO(&fds);
  FD_SET(sock, &fds);
  max_fd = sock;

  /* The main event loop. */

  while (1) {
    read_fds = fds;
    if (select(max_fd + 1, &read_fds, NULL, NULL, NULL) < 0) {
      perror("net_event_loop/select");
      /* Close all connections. */
      for (fd = 0; fd <= max_fd; fd++)
	if (FD_ISSET(fd, &fds))
	  close(fd);
      return;
    }

    if (FD_ISSET(sock, &read_fds)) {
      /* We have a new connection. */
      fd = accept(sock, NULL, NULL);
      if (fd == -1) {
	perror("net_event_loop/accept");
	continue;
      }
      if (fd > max_fd)
	max_fd = fd;
      FD_SET(fd, &fds);
    } else
      /* Incoming data from an existing connection. */
      for (fd = 0; fd <= max_fd; fd++)
	if (FD_ISSET(fd, &read_fds)) {
	  n_bytes = recv(fd, buf, BUF_SIZE, 0);
	  if (n_bytes <= 0) {
	    /* The connection closed.  Call callback with NULL so that
	       it might perform some clean ups. */
	    perror("net_event_loop/recv");
	    cb(fd, 0, NULL);
	    close(fd);
	    FD_CLR(fd, &fds);
	  } else {
	    /* Incoming data.  Add a terminating nil character in case
	       it helps the callback. */
	    buf[n_bytes + 1] = '\0';
	    /* Now perform the callback. */
	    if (cb(fd, n_bytes, buf) == 0) {
	      /* If the callback returned zero, close the connection. */
	      close(fd);
	      FD_CLR(fd, &fds);
	    }
	  }
	}
  }
}
