/* This file is part of toroidtoe, a web server implementing a game of
 * five-in-a-row on a grid warped onto a surface of a torus.  This
 * program is intended as a basis for student exercises on courses at
 * Helsinki University of Technology.  This source code is distributed
 * under the GPL.
 *
 * Author: Kenneth Oksanen <cessu@iki.fi>
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <strings.h>

#include "game.h"
#include "ai.h"


#ifndef FILENAME_MAX
/* FILENAME_MAX is defined in ISO C standard, and should be defined in
   stdio.h.  But some headers are not yet up to that, so define some
   FILENAME_MAX if it is not yet defined. */
#define FILENAME_MAX  1024
#endif


typedef struct session_t {
  long int id;
  int nth_move;
  position_t position;
  struct session_t *next;
} session_t;

static session_t *sessions = NULL;


static session_t *lookup_or_create_session(long int session_id)
{
  session_t *session;

  for (session = sessions; session != NULL; session = session->next)
    if (session->id == session_id)
      return session;
  /* Not found, create a new one. */
  session = calloc(1, sizeof(session_t));
  if (session == NULL) {
    fprintf(stderr, "malloc() failed.\n");
    return NULL;
  }
  session->id = session_id;
  /* Push this new session to the linked list of all sessions. */
  session->next = sessions;
  sessions = session;
  return session;
}


static long int new_session_id(void)
{
  return random() & 0xFFFFFFF;
}


static const char *graphics_dirname = NULL;

void session_init(const char *s)
{
  graphics_dirname = s;
}


/* Definitions of some HTTP reply kinds.  They are strongly
   abbreviated from what typical WWW servers would send and in fact
   they may therefore not work on all browsers.  Client-side caching
   has been disabled for all contents, partly because some pages are
   dynamic, and partly because caching may hide some changes made
   during development. */

const char gif_reply_hdr[] = "HTTP/1.1 200 OK\r\n\
Content-Type: image/gif\r\n\
Cache-control: no-cache\r\n\r\n";

const char html_reply_hdr[] = "HTTP/1.1 200 OK\r\n\
Content-type: text/html\r\n\
Cache-control: no-cache\r\n\r\n\
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\" \"http://www.w3.org/TR/REC-html40/loose.dtd\">\r\n";

const char first_page[] = "<HTML>\
  <BODY bgcolor=\"#FFFFFF\">\
    <TITLE>ToroidToe</TITLE>\
    <H1>ToroidToe</H1>\
    Welcome to ToroidToe, a game of five-in-a-row or noughts and crosses\
    on a 6-by-6 grid warped onto a surface of a torus.  The initiating player,\
    cross, has a clear advantage, and you will play it against the computer.\
    <P>\
    <A HREF=\".%07lX.0.0.0/\">Click to play</A>\
  </BODY>\
</HTML>";

const char not_found_reply[] = "HTTP/1.1 404 Not Found\r\n";

const char *allowed_images[] = {
	"cross-green.gif",
	"cross-grey.gif",
	"cross-white.gif",
	"cross.gif",
	"empty-grey.gif",
	"empty-white.gif",
	"not.gif",
	"nought-green.gif",
	"nought-grey.gif",
	"nought-white.gif",
	NULL};

/* Check if str is listed in a NULL terminated list. */

int is_listed(const char *str, const char **list) {
	while (*list) {
		if (!strcmp(*list, str))
			return 1;
		list++;
	}
	return 0;
}

/* The callback function called from networking code. */

int session_cb(int fd, int n_bytes, char *bytes)
{
  int i, x, y;
  FILE *fp = NULL;
  value_t value;
  session_t *session;
  long int session_id;
  char *s, filename[FILENAME_MAX + 1];

  assert(graphics_dirname != NULL);

  if (bytes == NULL)
    /* The connection dies, nothing to do. */
    return 0;

#if 0
  /* This may be useful during debugging.  I suggest you try it and notice
     that there's much more content in the request that we ignore. */
  fprintf(stderr, "HTTP request: %.*s\n", n_bytes, bytes);
#endif

  /* Parse the request to a sufficient extent. */

  /* We are interested only in the GET-requests. */
  if (strncmp(bytes, "GET ", 4) != 0)
    goto default_answer;
  bytes += 4;
  /* Terminate the request at the next space character.  Thereafter
     `bytes' should be the requested URI as a typical C string. */
  s = index(bytes, ' ');
  if (s == NULL)
    /* Request somehow corrupt. */
    goto default_answer;
  *s = '\0';
  
  /* URIs to gif images this program serves begin with "/pix/". */
  if (strncmp(bytes, "/pix/", 5) == 0) {
    if (strlen(graphics_dirname) + strlen(bytes + 4) > FILENAME_MAX)
      /* Too long to be a filename, someone's probably trying to crack. */
      goto default_answer;
    strcpy(filename, graphics_dirname);
    if (!is_listed(bytes+5, allowed_images))
      /* Attempt to request something else than images that belong to this project. */
      goto default_answer;
    strcat(filename, bytes + 4);
    fp = fopen(filename, "r");
    if (fp == NULL)
      /* Requested gif image wasn't found. */
      goto default_answer;
    write(fd, gif_reply_hdr, sizeof(gif_reply_hdr) - 1);
    while (!feof(fp)) {
      /* Reuse the filename buffer as temporary memory when copying
	 from fp to fd. */
      i = fread(filename, 1, FILENAME_MAX, fp);
      if (i < 0)
	/* Some kind of a problem, hopefully the user got the image. */
	break;
      write(fd, filename, i);
    }
    fclose(fp);
    return 0;
  }

  /* Convert the low-level file descriptor to a file pointer so that
     we can use for example fprintf to it.  We have to dup it so that
     we can close it correctly (without closing it a second time, or
     leaking FILE's). */
  i = dup(fd);
  if (i == -1)
    /* Most unlikely error, but might happen, e.g., under heavy load. */
    goto default_answer;
  fp = fdopen(i, "a");
  if (fp == NULL)
    goto default_answer;

  /* The first page. */
  if (strcmp(bytes, "/") == 0) {
    fprintf(fp, "%s", html_reply_hdr);
    fprintf(fp, first_page, new_session_id());
    fclose(fp);
    return 0;
  }

  /* The game in progress.  We use the hexadecimal value as a
     probabilistic indicator of which game is in progress. */
  if (strncmp(bytes, "/.", 2) != 0)
    goto default_answer;
  bytes += 2;
  session_id = strtol(bytes, &bytes, 16);
  session = lookup_or_create_session(session_id);
  if (session == NULL)
    goto default_answer;
  /* Continue parsing the move number.  Check that the player does not
     undo. */
  if (*bytes != '.')
    goto default_answer;
  i = strtol(bytes + 1, &bytes, 10);
  if (*bytes != '.')
    goto default_answer;
  x = strtol(bytes + 1, &bytes, 10);
  if (*bytes != '.')
    goto default_answer;
  y = strtol(bytes + 1, &bytes, 10);

  if (i == session->nth_move + 1) {
    /* The user has clicked on a link which indicates a new move to
       x,y.  Compute our reply and increment move counter. */
    if (x < 0 || x >= NX || y < 0 || y >= NY)
      goto default_answer;
    session->position[x][y] = CROSS;
    if (!game_is_over(session->position, &value)) {
      ai_move(session->position, NOUGHT, &x, &y, 1);
      session->position[x][y] = NOUGHT;
    }
    session->nth_move++;
  }

  /* Produce a HTML table that shows the current position and as
     active links. */
  fprintf(fp, "%s<HTML>\
  <BODY bgcolor=\"#FFFFFF\">\
    <TITLE>ToroidToe</TITLE>\
    <H1>ToroidToe, move %d</H1>", html_reply_hdr, session->nth_move);
  if (i != session->nth_move)
    fprintf(fp, "You can't undo moves.<P>\n");
  fprintf(fp, "<TABLE>");
  for (x = 0; x < NX; x++) {
    fprintf(fp, "<TR>");
    for (y = 0; y < NY; y++)
      fprintf(fp,
	      "<TD><A HREF=\"/.%07lX.%d.%d.%d\">\
    <IMG SRC=\"/pix/%s-white.gif\">\
  </A>\n",
	      session_id, i + 1, x, y, 
	      session->position[x][y] == EMPTY ? "empty" :
	      session->position[x][y] == CROSS ? "cross" : "nought");
  }
  fprintf(fp, "</TABLE>");
  if (game_is_over(session->position, &value))
    fprintf(fp, "<P>%s<P><A HREF=\".%07lX.0.0.0/\">New game</A>",
	    value == WIN ? "Congratulations, you won" :
	    value == -WIN ? "I won!" : "The game ended in a draw",
	    new_session_id());
  fprintf(fp, "</BODY></HTML>");

  fclose(fp);
  return 0;

 default_answer:
  if (fp != NULL)
    fclose(fp);

  /* Default reply is that the requested URL was not found. */
  write(fd, not_found_reply, sizeof(not_found_reply) - 1);

  return 0;
}
